package cubeengine;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics2D;
import java.io.Serializable;
import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Figure-objekt r figurer som kan renderas
 * av CubeEngine
 * 
 * @author Per-Erik Kristensson
 *
 */
public class Figure implements Serializable
{
	/**
	 * Skapar en tom Figure
	 */
	public Figure()
	{
	}

	/**
	 * Skapar en Figure genom att lsa in en Figure
	 * frn en Xml-fil.
	 * 
	 * @param s Namnet p figuren som ska lsas.
	 */
	public Figure(String s)
	{
		load(s);
	}

	/**
	 * Retunerar Figurens mittkordat.
	 * @return Figurens mittkordat.
	 */
	public Cord3d getCenter()
	{
		return m_center;
	}
	
	/**
	 * Stter Figurens mittkordinat
	 * @param center Figurens nya centrum
	 */
	public void setCenter(Cord3d center)
	{
		m_center = center;
	}

	/**
	 * Normaliserar figuren. Gr att medelvrdet av alla punkter r (0,0,0),
	 * och stter centrum till denna punkt.
	 */
	public void normalize()
	{
		m_center = Cord3d.getCenter( m_cords );

		//Fungerar nog bara om figuren r symetrisk... eller?
		for(int i=0; i<getCordCount(); i++)
			getCord(i).subtract( m_center );

		m_center = new Cord3d(0,0,0);
	}

	/**
	 * Stter frgen p alla sidor
	 * @param color Sidfrg
	 */
	public void setAllSideColor(Color color)
	{
		for(int s=0; s<getSideCount(); s++)
			getSide(s).setColor(color);
	}

	/**
	 * Stter basfrgen p alla sidor
	 * @param color Sidbasfrg
	 */
	public void setAllSideBaseColor(Color color)
	{
		for(int s=0; s<getSideCount(); s++)
			getSide(s).setBaseColor(color);
	}

	/**
	 * Stter ramfrgen p alla sidor
	 * @param color Ramfrg
	 */
	public void setAllSideBorderColor(Color color)
	{
		for(int s=0; s<getSideCount(); s++)
			getSide(s).setBorderColor(color);
	}
	
	/**
	 * Stter basramfrgen p alla sidor
	 * @param color Basramfrg
	 */
	public void setAllSideBorderBaseColor(Color color)
	{
		for(int s=0; s<getSideCount(); s++)
			getSide(s).setBorderBaseColor(color);
	}

	/**
	 * @param n Sidnummer
	 * @return sida nummer n
	 */
	public Side getSide(int n)
	{
		return (Side) m_sides.get(n);
	}

	/**
	 * 
	 * @param n Kordinatnummer
	 * @return kordinat nummer n
	 */
	public Cord3d getCord(int n)
	{
		return (Cord3d) m_cords.get(n);
	}       

	/**
	 * Lgger till en sida
	 * @param s sida som ska lggas till
	 */
	public void addSide( Side s)
	{
		m_sides.add( s );
	}

	/**
	 * Lgger till en kordinat
	 * @param c Kordinat som ska lggas till
	 */
	public void addCord( Cord3d c)
	{
		m_cords.add( c );
	}

	/**
	 * Retunerar antalet sidor
	 * @return antalet sidor
	 */
	public int getSideCount()
	{
		return m_sides.size();
	}

	/**
	 * Retunerar antalet kordinater
	 * @return antalet kordinater
	 */
	public int getCordCount()
	{
		return m_cords.size();
	}

	/**
	 * Skalar om figuren
	 * @param x Omskalning i x-led
	 * @param y Omskalning i y-led
	 * @param z Omskalning i z-led
	 */
	public void rescale(double x, double y, double z)
	{
		for(int i=0; i<getCordCount(); i++)
		{
			Cord3d c = getCord(i);

			c.set( x*c.getX(), y*c.getY(), z*c.getZ() );
		}
	}

	/**
	 * Letar upp en nod
	 * @param list		Lista att ska i
	 * @param tagid		tag som ska skas efter
	 * @param attribute Vilket attribut som taggen mste ha. Kan vara null 
	 * @param value		Vilket vrde attributet mste ha. Kan vara null
	 * @return Frsta noden som r av typen tagid, och matchande attribut
	 */
	protected static Node findNode(NodeList list, String tagid, String attribute, String value)
	{
		if(list == null)
			return null;

		for(int i=0; i<list.getLength(); i++)
		{
			Node n = list.item(i);

			if( tagid.compareTo( n.getNodeName() ) == 0)
			{
				//Kolla inte attribut
				if(attribute == null || value == null)
					return n;

				String nodevalue = getAttribute(n, attribute);

				//Matchar vrdet p attributet?
				if(nodevalue != null && value.equals(nodevalue))
					return n;
			}
		}

		return null;
	}

	/**
	 * Hmtar vrdet p ett attribut i en tagg
	 * @param n Nod att hmta vrde ifrn
	 * @param attribute Vilket attribut som ska hmtas
	 * @return Vrde p attributet
	 */
	protected static String getAttribute(Node n, String attribute)
	{
		NamedNodeMap map = n.getAttributes();
		Node n2=map.getNamedItem(attribute);

		if(n2 == null)
			return null;

		return n2.getNodeValue();
	}
	
	/**
	 * Kollar om en punkt r inne i figuren
	 * @param x  Punkt, i x-led
	 * @param y  Punkt, i y-led
	 * @param z  Punkt, i z-led
	 * @return true om punkten r i figuren
	 */
	public boolean isInside(double x, double y, double z)
	{
		//Kollar om punkten r innanfr figurens radie.
		//Vore trevligt om det vore mer exakt.
		if( m_center.getDistance(x,y,z) <= m_radius ) 
			return true;		
		
		return false;
	}

	/**
	 * Kollar om en kordinat r inne i figuren
	 * @param p Kordinat
	 * @return true om kordinaten r inne i figuren
	 */
	public boolean isInside(Cord3d p)
	{
		return isInside(p.getX(), p.getY(), p.getZ());
		
	}
	
	/**
	 * Kollar om en figure r i denna figure 
	 * @param f Figure
	 * @return true om Figur f r i denna figur
	 */
	public boolean isInside(Figure f)
	{
		double dx = getCenter().getX() - f.getCenter().getX();
		double dy = getCenter().getY() - f.getCenter().getY();
		double dz = getCenter().getZ() - f.getCenter().getZ();
		
		double r = getRadius() + f.getRadius();
		
		return dx*dx+dy*dy+dz*dz < r*r;
	}

	/**
	 * Lser in kordinater frn en xml-nod 
	 * @param points XML-nod
	 * @return en ArrayList med alla kordinater
	 */
	protected ArrayList loadCords( Node points )
	{
		double xscale=1, yscale=1, zscale=1;

		//Ls in skalning
		{
			String s = getAttribute(points, "xscale");
			if(s!=null) xscale = Double.parseDouble(s);

			s = getAttribute(points, "yscale");
			if(s!=null) yscale = Double.parseDouble(s);

			s = getAttribute(points, "zscale");
			if(s!=null) zscale = Double.parseDouble(s);                   
		}


		int i=0;
		ArrayList cords = new ArrayList();

		while(true)
		{
			Node point=findNode(points.getChildNodes(), "point", "num", ""+i);

			if(point == null)
				break;

			double x = Double.parseDouble( getAttribute(point, "x") ) * xscale;
			double y = Double.parseDouble( getAttribute(point, "y") ) * yscale;
			double z = Double.parseDouble( getAttribute(point, "z") ) * zscale;

			Cord3d cord = new Cord3d(x, y, z);
			cords.add(cord);

			i++;
		}

		return cords;
	}

	/**
	 * Lser in frgen frn en XML-nod
	 * @param n Nod att lsa ifrn
	 * @param defaultcolor Standardfrg
	 * @return frgen i noden
	 */
	protected Color getColor( Node n, Color defaultcolor )
	{
		if(n != null)
		{
			int r = Integer.parseInt( getAttribute(n, "r") );
			int g = Integer.parseInt( getAttribute(n, "g") );
			int b = Integer.parseInt( getAttribute(n, "b") );

			return new Color(r, g, b);                        
		}

		return defaultcolor;
	}

	/**
	 * Lser in sidor frn en nod
	 * @param sides XML-nod med alla sidor 
	 * @param cords ArrayList med alla punkter sidorna lnkas till
	 * @return en ArrayList med alla sidor.
	 */
	protected ArrayList loadSides( Node sides, ArrayList cords )
	{
		int i=0;

		ArrayList sidevec = new ArrayList();

		Color defcolor	  = getColor( findNode( sides.getChildNodes(), "defaultcolor", "type", "color"), null );
		Color defbasecolor = getColor( findNode( sides.getChildNodes(), "defaultcolor", "type", "basecolor"), null );

		Color defborderbasecolor = getColor( findNode( sides.getChildNodes(), "defaultcolor", "type", "borderbasecolor"), null );
		Color defbordercolor = getColor( findNode( sides.getChildNodes(), "defaultcolor", "type", "bordercolor"), null );

		//Ls in alla sidor
		while(true)
		{
			Node sidenode=findNode( sides.getChildNodes(), "side", "num", ""+i);

			if(sidenode == null) break;

			Color sidecolor = getColor( findNode( sidenode.getChildNodes(), "color", "type", "color"), defcolor ); 

			Color basecolor = getColor( findNode( sidenode.getChildNodes(), "color", "type", "basecolor"), defbasecolor );
	
			Color bordercolor = getColor( findNode( sidenode.getChildNodes(), "color", "type", "bordercolor"), defbordercolor );
			
			Color borderbasecolor = getColor( findNode( sidenode.getChildNodes(), "color", "type", "borderbasecolor"), defborderbasecolor );


			int c=0;
			ArrayList sidecords = new ArrayList();

			while(true)
			{
				Node corner = findNode( sidenode.getChildNodes(), "corner", "num", ""+c );

				if(corner == null) break;

				int cordid = Integer.parseInt( getAttribute(corner, "point" ) );
				sidecords.add( (Cord3d) cords.get(cordid) );
    
				c++;
			}

			Cord3d[] cs = new Cord3d[ sidecords.size() ];
			sidecords.toArray(cs);

			Side side = new Side(cs, sidecolor, this);
			
			side.m_baseColor = basecolor;
			side.m_borderColor = bordercolor;
			side.m_borderBaseColor = borderbasecolor;


			sidevec.add( side );

			i++;
		}

		return sidevec;
	}

	/**
	 * Hmtar alla namn p figurer i XML-filen "figures.xml"
	 * @return en array med alla namn i XML-filen
	 */
	public static String[] getAllFigureNames()
	{
		DocumentBuilderFactory factory =
			DocumentBuilderFactory.newInstance();

		String[] figNames=null;

		try
		{
			DocumentBuilder builder = factory.newDocumentBuilder();
			Document document = builder.parse( "figures.xml" );

			Element root = document.getDocumentElement();

			NodeList figures = root.getChildNodes();

			ArrayList names = new ArrayList();

			for(int i=0; i<figures.getLength(); i++)
			{
				Node n = figures.item(i);

				if( n.getNodeName().compareTo( "figure" ) == 0)
				{
					names.add( getAttribute(n, "name") );
				}
			}

			figNames = new String[ names.size() ];
			names.toArray(figNames);
            

		}
		catch(Exception e)
		{
			return null;
		}

		return figNames;

	}

	/**
	 * Lser in en figur frn XML-filen
	 * @param figname Namnet p figuren
	 */
	public void load(String figname)
	{

		DocumentBuilderFactory factory =
			DocumentBuilderFactory.newInstance();

		try 
		{
			DocumentBuilder builder = factory.newDocumentBuilder();
			//Document document = builder.parse( "figures.xml" );
			
			Document document = null;
			InputStream is = getClass().getResourceAsStream("figures.xml" );
			if(is != null)
			{
			  document = builder.parse( is );
			}
			else
			{
			  try
			  {
			   document = builder.parse( "figures.xml" );
			  }
			  catch(Exception e)
			  {
			    return;
			  }
			}


			Element root = document.getDocumentElement();

			NodeList children = root.getChildNodes();

			Node fig=findNode(children, "figure", "name", figname);

			Node settings=findNode( fig.getChildNodes(), "settings", null, null );

			if(settings!=null)
			{
				String s = getAttribute(settings, "radius");
				if(s!=null) m_radius = Double.parseDouble(s);

				s = getAttribute(settings, "maxHidden");
				if(s!=null) m_maxHiddenSideCount = Integer.parseInt(s);

				s = getAttribute(settings, "transparency");
				if(s!=null)	m_transparency = Float.parseFloat(s);
				
			}


			ArrayList cords = loadCords( findNode(fig.getChildNodes(), "points", null, null) );

			ArrayList sides = loadSides( findNode(fig.getChildNodes(), "sides", null, null), cords );

			if(cords != null && sides != null)
			{
				m_cords = cords;
				m_sides = sides;
			}

			normalize();
			

		}
		catch (Exception pce) 
		{
			// Parser with specified options can't be built
			pce.printStackTrace();

		}
	}

	/**
	 * Mter avstndet till en figure
	 * @param f Figure att mta avstnd till
	 * @return Avstndet till Figure f
	 */
	public double getDistance(Figure f)
	{
		return m_center.getDistance(f);
	}

	/**
	 * Roterar figuren
	 * @param matrix Rotationsmatris
	 */
	public void rotate(double[][] matrix)
	{
		m_center.rotate( matrix );
	}

	/**
	 * Roterar figuren
	 * @param vx Vinkel i x-led
	 * @param vy Vinkel i y-led
	 * @param vz Vinkel i z-led
	 */
	public void rotate(double vx, double vy, double vz)
	{
		m_center.rotate( Cord3d.createRotateMatrix(vx, vy, vz) );
	}

	/**
	 * Sorterar sidorna s att de sidor som r nrmast Cameran camera
	 * hamnar sist
	 * 
	 * @param camera
	 */
	protected void sortSides(Camera camera)
	{
		class SideComparer implements Comparator 
		{
			SideComparer(Camera c)
			{
				m_camera = c;
			}

			public int compare(Object obj1, Object obj2)
			{
				Side s1 = (Side) obj1;
				Side s2 = (Side) obj2;

				if( m_camera.getDistance(s1) < m_camera.getDistance(s2) )
					return 1;

				return -1;
			}

			Camera m_camera;
		}


		//Se till att centrum p sidan r relativ figurens centrum
		for(int i=0; i<m_sides.size(); i++)
		{
			( (Side) m_sides.get(i) ).recalcCenter();

			( (Side) m_sides.get(i) ).getCenter().add(m_center);
		}


		//Sortera sidorna s att de att sidor lngst frn kameran hamnar frst.
		Collections.sort( m_sides, new SideComparer(camera) );

	}

	/**
	 * Ritar upp figuren
	 * @param g	Graphics2d att rita i
	 * @param centerx Mittpunkt i xled
	 * @param centery Mittpunkt i yled
	 * @param sun Solar
	 * @param camera Kamera som ser p figuren
	 */
	public void draw(Graphics2D g, int centerx, int centery, Sun[] sun, Camera camera)
	{
		//Cacha
		if( !recalcVisualCords(camera, centerx, centery) )
			return;

		sortSides(camera);

		drawSetTransparency(g);

		for(int i=m_maxHiddenSideCount; i<m_sides.size(); i++)
			( (Side) m_sides.get(i) ).draw(g, centerx, centery, sun, camera);      
 
		drawResetTransparency(g);
	}

	/**
	 * Stter transparansen i Graphics2d. Ett anrop p denna
	 * mste fljas av ett anrop p drawResetTransparency.
	 * 
	 * @param g Objekt att stta transparans i 
	 */
	protected void drawSetTransparency(Graphics2D g)
	{
		if(m_transparency < 1.0f)
		{			 
			g.setComposite( AlphaComposite.getInstance(AlphaComposite.SRC_ATOP, m_transparency) );
		}
	}

	/**
	 * terstller transparansen i Graphics2d
	 * @param g Objekt som ska terstllas
	 */
	protected void drawResetTransparency(Graphics2D g)
	{
		if(m_transparency < 1.0f)
		{
			g.setComposite( AlphaComposite.getInstance(AlphaComposite.SRC_ATOP) );
		}
	}

	/**
	 * Roterar figuren runt sin egen axel
	 * @param matrix Rotationsmatris
	 */
	public void rotateAxel(double[][] matrix)
	{    
		//Rkna ut de nya punkterna.
		for(int i=0; i<getCordCount(); i++)
		{
			double newx = getCord(i).getX() * matrix[0][0] + getCord(i).getY() *matrix[0][1] + getCord(i).getZ() *matrix[0][2];
			double newy = getCord(i).getX() * matrix[1][0] + getCord(i).getY() *matrix[1][1] + getCord(i).getZ() *matrix[1][2];
			double newz = getCord(i).getX() * matrix[2][0] + getCord(i).getY() *matrix[2][1] + getCord(i).getZ() *matrix[2][2];

			getCord(i).set(newx, newy, newz);
		}
	}

	/**
	 * Roterar figuren runt sin en egen axel
	 * @param vx Vinkel i x-led
	 * @param vy Vinkel i y-led
	 * @param vz Vinkel i z-led
	 */
	public void rotateAxel(double vx, double vy, double vz)
	{    
		rotateAxel( Cord3d.createRotateMatrix(vx,vy,vz) );
	}

	/**
	 * Rknar ut pixelkordinaterna i alla kordinater, och lagrar vrdet
	 * i cache-delen.
	 * @param camera Kameran som tittar p figuren
	 * @param centerx Mittpunkt i x-led
	 * @param centery Mittpunkt i y-led
	 * @return false om inga punkter r synliga
	 */
	protected boolean recalcVisualCords(Camera camera, int centerx, int centery)
	{
		int minx=0, miny=0, maxx=0, maxy=0;

		for(int i=0; i<getCordCount(); i++)
		{
			//Ta fram den riktiga punkten i rummet
			//(kordinaterna r relativ kuben)
			Cord3d realCord = null; 
    
			if(m_parent!=null)
				realCord = Cord3d.sum(getCord(i), m_parent.m_center);
			else
				realCord = Cord3d.sum(getCord(i), m_center);
			
			//Bakom kameran?
			//if(camera.getCameraZ(realCord) >= -0.004)
			//	return false;

			Cord3d relCord = camera.getCameraCord(realCord);

			int x = getCord(i).m_cache_visualx =  camera.getVisualXPos(realCord, centerx);
			int y = getCord(i).m_cache_visualy =  camera.getVisualYPos(realCord, centery);

			if(i==0)
			{
				minx=maxx=x;
				miny=maxy=y;
			}
			else
			{
				if(x > maxx)
					maxx=x;
				else if(x < minx)
					minx=x;

				if(y > maxy)
					maxy=y;
				else if(y < miny)
					miny=y;
			}
		}

		//Alla punkter utanfr omrdet?
		if( minx > 2*centerx || maxx < 0 ||
			miny > 2*centery || maxy < 0)
		{
			//Rita inte ut!
			return false;
		}

		return true;
	}

	/**
	 * Hmtar figurens frlder
	 * @return Figuren frlder
	 */
	protected Figure getParent()
	{
		return m_parent;
	}

	/**
	 * Stter figurens frlder
	 * @param parent ny frlder
	 */
	public void setParent(Figure parent)
	{
		m_parent = parent;
	}

	/**
	 * Hmtar figurens radie
	 * @return Figurens radie
	 */
	public double getRadius()
	{
		return m_radius;
	}

	/**
	 * Stter figurens radie
	 * @param r Figurens radie
	 */
	public void setRadius(double r)
	{
		m_radius = r;
	}

	/**
	 * Kollar om figuren r synlig
	 * @return true om figuren r synlig
	 */
	public boolean isVisible()
	{
		return m_visible;
	}
	
	/**
	 * Bestmmer om figuren r synlig
	 * @param visible nytt vrde
	 */
	public void setVisible(boolean visible)
	{
		m_visible=visible;
	}

	/**
	 * Stter figurens transparans
	 * @param t Nytt vrde
	 */
	public void setTransparency(float t)
	{
		m_transparency=t;
	}

	/**
	 * Hmtar figurens transparans
	 * @return figurens transparans
	 */
	public float getTransparency()
	{
		return m_transparency;
	}
	
	
	/**
	 * Kordinat som motsvarar mitten p figuren
	 */
	protected Cord3d m_center = new Cord3d(0,0,0);	

	/**
	 * Array med alla sidor
	 */
	protected ArrayList m_sides = new ArrayList();
	
	/**
	 * Array med alla hrn
	 */
	protected ArrayList m_cords = new ArrayList();

	/**
	 * Om figuren r synlig eller ej
	 */
	protected boolean m_visible=true;
	
	/**
	 * Figurens transparans (r nog opacitet egentligen)
	 */
	protected float m_transparency=1.0f;
	
	/**
	 * Figurens frlder
	 */
	protected Figure m_parent=null;

	/**
	 * Hur mnga sidor som maximalt kan vara dolda vid uppritning
	 */
	protected int m_maxHiddenSideCount = 0;

	/**
	 * Figurens radie.
	 */
	protected double m_radius=0;
}

