package cubeengine;

import java.awt.Color;
import java.awt.Graphics2D;
import java.io.Serializable;

/**
 * En sida antas vara en platt yta. Sidan bestr av flera olika punkter,
 * och har fyra frger. Sidan har en basfrg som r minsta vrdet 
 * p frgen sidan kan ha. Sidan har ocks en "riktig" frg som kommer
 * adderas p basfrgen beroende p hur sidan blir belyst av en sol.
 * r basfrgen null ritas en frgen ut. De andra tv frgen r p sidan kant,
 * de fungerar p samma stt som frgena p ytan.
 * 
 * @author Per-Erik Kristensson
 */
public class Side extends Figure implements Serializable
{
    /**
     * Skapar en tom sida.
     */
	public Side(){}

	/**
	 * Skapar en sida med fyra hrn
	 * 
	 * @param c1 Hrn 1
	 * @param c2 Hrn 2
	 * @param c3 Hrn 3
	 * @param c4 Hrn 4
	 * @param color Sidan frg
	 * @param parent Sidans frlder
	 */
	public Side(Cord3d c1, Cord3d c2, Cord3d c3, Cord3d c4, Color color, Figure parent)
	{
		addCord(c1);
		addCord(c2);
		addCord(c3);
		addCord(c4);

		m_color = color;
		setParent( parent );
		
		calcMaxCornerDiffLength();
	}

	/**
	 * Skapar en sida med ett antal hrn
	 * @param cords Punkterna som hrn
	 * @param color Sidans frg
	 * @param parent Sidans frlder
	 */
	public Side(Cord3d[] cords, Color color, Figure parent)
	{
		for(int i=0; i<cords.length; i++)
			addCord( cords[i] );

		m_color = color;
		setParent( parent );
		
		calcMaxCornerDiffLength();
	}

	/**
	 * Stter figurens centrum som medelvrdet av alla hrnpunkter	 
	 */
	public void recalcCenter()
	{
		getCenter().set( Cord3d.getCenterX(m_cords), Cord3d.getCenterY(m_cords), Cord3d.getCenterZ(m_cords) );
	}

	/**
	 * Kollar om en sida r framfr en annan sida
	 * @param c Sida att jmfra med
	 * @return true om denna sida r framfr en annan sida
	 */
	public boolean isInFrontOf(Side c)
	{
		return getCenter().getZ() > c.getCenter().getZ();
	}

	/**
	 * Rknar ut det maximala avstndet mellan tv punkter.
	 * Denna MSTE anropas nr man har lagt till alla punkter
	 * i sidan, annars kan uppritningen g fel.
	 */
	public void calcMaxCornerDiffLength()
	{
	    m_maxCornerDiffLength = 0;
	    
	    for(int i=0; i<getCordCount()-1; i++)
	        for(int j=i+1; j<getCordCount(); j++)
	        {
	            double diff = getCord(i).getDistance(getCord(j));
	            
	            //Rekortlngd?
	            if(diff > m_maxCornerDiffLength)
	                m_maxCornerDiffLength=diff;
	        }
	}
	

	/**
	 * Ritar ut en sida.
	 */
	public void draw(  Graphics2D g, 
		int centerx, 
		int centery, 
		Sun[] allsun,
		Camera camera)
	{     
		Color sideColor = m_color;

		int[] currentSideRGB = { 0, 0, 0 };
		if(m_baseColor != null)	getRGB(currentSideRGB, m_baseColor);

		int[] sideColorRGB = { 0, 0, 0 };
		if(sideColor != null) getRGB(sideColorRGB, sideColor);


		int[] currentBorderRGB = { 0, 0, 0 };
		if(m_borderBaseColor != null)	getRGB(currentBorderRGB, m_borderBaseColor);

		int[] sideBorderRGB = { 0, 0, 0 };
		if(m_borderColor != null) getRGB(sideBorderRGB, m_borderColor);


		if(getParent() != null)
		{
			recalcCenter();

			getCenter().add(getParent().getCenter());
		}

		//Cacha om enskild sida (annars har frlderna redan gjort det)
		if(getParent() == null)
		{
			if( !recalcVisualCords(camera, centerx, centery) )
				return;

			drawSetTransparency(g);
		}

		//Ta fram kordinaterna p skrmen fr alla punkter
		int[] x = new int[getCordCount()];
		int[] y = new int[getCordCount()];

		for(int i=0; i<getCordCount(); i++)
		{
			x[i] = getCord(i).m_cache_visualx;
			y[i] = getCord(i).m_cache_visualy;
		}

		//Rkna om frgen sidan med hjlp av belysning
		if(allsun != null && (m_color!=null || m_borderColor!=null) )
			for(int ww=0; ww<allsun.length; ww++)
			{
				Sun sun = allsun[ww];

				//Rkna ut avstndet till solen
				double centerdistance = sun.getDistance( this );

				if(getParent() != null)
				{
					double cubedistance = sun.getDistance( getParent() );

					//Sidan i skugga av kuben?
					if(centerdistance >= cubedistance)
						continue;
				}

				double gamma = 1.0;

				//Hur mycket frgfrndring?
				{
					double lowDiff=0, hiDiff=0;

					//Ta fram de punkter som r lngst och nrmast solen.
					for(int i=0; i<getCordCount(); i++)
					{
						double diff = getCord(i).getDistance( sun );

						if(i==0)
							hiDiff=lowDiff=diff;
						else
						{
							if(diff>hiDiff)
								hiDiff=diff;
							else if(diff < lowDiff)
								lowDiff=diff;
						}
					}

					//Gamma r denna kvot
					gamma = //Math.sqrt
						(1-(hiDiff-lowDiff)/m_maxCornerDiffLength);
            
					if(gamma<0)
						continue;
            
					//Frsvaga om avstnd lngre n 2
					if(centerdistance > 2)
						gamma*= 2/(centerdistance);

				}
    
				gamma *= sun.getGamma();

				//Ta fram frger p solen
				int[] sunRGB = { sun.getColor().getRed(), sun.getColor().getGreen(), sun.getColor().getBlue() };

				//Berkna de nya frgerna
				if(sideColor != null)
					modifyColor(currentSideRGB, sideColorRGB, sunRGB, gamma);

				if(m_borderColor != null)
					modifyColor(currentBorderRGB, sideBorderRGB, sunRGB, gamma);                        
				    
			}


		//Rita ut bas...
		if(m_baseColor != null)
		{
			g.setColor( new Color( currentSideRGB[0], currentSideRGB[1], currentSideRGB[2]));
			g.fillPolygon(x, y, x.length);
		}

		//...och kant
		if(m_borderBaseColor != null)
		{
			g.setColor( new Color( currentBorderRGB[0], currentBorderRGB[1], currentBorderRGB[2]) );
			g.drawPolygon(x, y, x.length);
		}


		if(getParent() == null)
		{
			drawResetTransparency(g);
		}

	}

	/**
	 * Omvandlar en color till en vektor
	 * @param rgb vektor att lagra vrdena i
	 * @param color Frgen att omvandla
	 */
	static protected void getRGB(int[] rgb, Color color)
	{
		rgb[0] = color.getRed();
		rgb[1] = color.getGreen();
		rgb[2] = color.getBlue();
	}

	/**
	 * Frndrar en frg
	 * @param currentColor En vektor med de nuvarande rgb-frger
	 * @param baseColor En vektor med de nuvarande rgb-frgerna fr basfrgen
	 * @param sunColor En vektor med de nuvarande rgb-frgerna fr solen
	 * @param gamma "Gamma"vrdet
	 */
	static protected void modifyColor(int[] currentColor, int[] baseColor, int[]sunColor, double gamma)
	{
		for(int i=0; i<3; i++)
		{
			currentColor[i] += (double)baseColor[i]*(gamma*sunColor[i]/255.0);
            
			if(currentColor[i] > 255)
				currentColor[i] = 255;
		}                   
	}

	/**
	 * Hmtar sidan frg
	 * @return sidans frg
	 */
	public Color getColor()
	{
		return m_color;
	}

	/**
	 * Hmtar sidans basfrg
	 * @return sidans basfrg
	 */
	public Color getBaseColor()
	{
		return m_baseColor;
	}

	/**
	 * Hmtar sidans kantfrg
	 * @return sidans kantfrg
	 */
	public Color getBorderColor()
	{
		return m_borderColor;
	}
	
	/**
	 * Hmtar sidans kantbasfrg
	 * @return sidans kantbasfrg
	 */
	public Color getBorderBaseColor()
	{
		return m_borderBaseColor;
	}

	/**
	 * Stter sidans frg
	 * @param c Ny frg
	 */
	public void setColor(Color c)
	{
		m_color=c;
	}
	
	/**
	 * Stter sidans basfrg
	 * @param c Ny frg
	 */
	public void setBaseColor(Color c)
	{
		m_baseColor=c;
	}

	/**
	 * Stter sidans kantfrg
	 * @param c Ny frg
	 */
	public void setBorderColor(Color c)
	{
		m_borderColor=c;
	}

	/**
	 * Stter sidans kantbasfrg
	 * @param c Ny frg
	 */
	public void setBorderBaseColor(Color c)
	{
		m_borderBaseColor=c;
	}
	
	/**
	 * Sidan frg och basfrger
	 */
	protected Color
		/**
		 * Sidans frg
		 */
		m_color,
		/**
		 * Sidans basfrg
		 */
		m_baseColor = new Color(8,8,8),
		/**
		 * Kantens frg
		 */
		m_borderBaseColor=null,
		/**
		 * Kantens basfrg
		 */
		m_borderColor=null;
	
	/**
	 * Det lngsta avstndet mellan tv hrn
	 */
	protected double m_maxCornerDiffLength=0;
}

